import asyncio
import json
import logging
from pathlib import Path
from typing import Any

import aiohttp.web
import aiohttp_middlewares.cors

from pydase.config import ServiceConfig, WebServerConfig
from pydase.data_service.data_service_observer import DataServiceObserver
from pydase.server.web_server.sio_setup import (
    setup_sio_server,
)
from pydase.utils.helpers import (
    get_path_from_path_parts,
    parse_full_access_path,
)
from pydase.utils.serialization.serializer import generate_serialized_data_paths

logger = logging.getLogger(__name__)
API_VERSION = "v1"


class WebServer:
    """
    Represents a web server that adheres to the AdditionalServerProtocol, designed to
    work with a DataService instance. This server facilitates client-server
    communication and state management through web protocols and socket connections.

    The WebServer class initializes and manages a web server environment using FastAPI
    and Socket.IO, allowing for HTTP and WebSocket communications. It incorporates CORS
    (Cross-Origin Resource Sharing) support, custom CSS, and serves a frontend static
    files directory. It also initializes web server settings based on configuration
    files or generates default settings if necessary.

    Configuration for the web server (like service configuration directory and whether
    to generate new web settings) is determined in the following order of precedence:
    1. Values provided directly to the constructor.
    2. Environment variable settings (via configuration classes like
      `pydase.config.ServiceConfig` and `pydase.config.WebServerConfig`).
    3. Default values defined in the configuration classes.

    Args:
        data_service_observer (DataServiceObserver): Observer for the DataService,
          handling state updates and communication to connected clients.
        host (str): Hostname or IP address where the server is accessible. Commonly
          '0.0.0.0' to bind to all network interfaces.
        port (int): Port number on which the server listens. Typically in the range
          1024-65535 (non-standard ports).
        css (str | Path | None, optional): Path to a custom CSS file for styling the
          frontend. If None, no custom styles are applied. Defaults to None.
        enable_cors (bool, optional): Flag to enable or disable CORS policy. When True,
          CORS is enabled, allowing cross-origin requests. Defaults to True.
        config_dir (Path | None, optional): Path to the configuration
          directory where the web settings will be stored. Defaults to
          `pydase.config.ServiceConfig().config_dir`.
        generate_new_web_settings (bool | None, optional): Flag to enable or disable
          generation of new web settings if the configuration file is missing. Defaults
          to `pydase.config.WebServerConfig().generate_new_web_settings`.
        **kwargs (Any): Additional unused keyword arguments.
    """

    def __init__(  # noqa: PLR0913
        self,
        data_service_observer: DataServiceObserver,
        host: str,
        port: int,
        css: str | Path | None = None,
        enable_cors: bool = True,
        config_dir: Path = ServiceConfig().config_dir,
        generate_web_settings: bool = WebServerConfig().generate_web_settings,
        frontend_src: Path = Path(__file__).parent.parent.parent / "frontend",
    ) -> None:
        self.observer = data_service_observer
        self.state_manager = self.observer.state_manager
        self.service = self.state_manager.service
        self.port = port
        self.host = host
        self.css = css
        self.enable_cors = enable_cors
        self.frontend_src = frontend_src
        self._service_config_dir = config_dir
        self._generate_web_settings = generate_web_settings
        self._loop: asyncio.AbstractEventLoop
        self._initialise_configuration()

    async def serve(self) -> None:
        self._loop = asyncio.get_running_loop()
        self._sio = setup_sio_server(self.observer, self.enable_cors, self._loop)

        async def index(request: aiohttp.web.Request) -> aiohttp.web.FileResponse:
            return aiohttp.web.FileResponse(self.frontend_src / "index.html")

        app = aiohttp.web.Application()

        # Add CORS middleware if enabled
        if self.enable_cors:
            app.middlewares.append(
                aiohttp_middlewares.cors.cors_middleware(allow_all=True)
            )

        # Define routes
        self._sio.attach(app, socketio_path="/ws/socket.io")
        app.router.add_static("/assets", self.frontend_src / "assets")
        app.router.add_get("/service-properties", self._service_properties_route)
        app.router.add_get("/web-settings", self._web_settings_route)
        app.router.add_get("/custom.css", self._styles_route)

        app.router.add_get(r"/", index)
        app.router.add_get(r"/{tail:.*}", index)

        await aiohttp.web._run_app(
            app,
            host=self.host,
            port=self.port,
            handle_signals=False,
            print=logger.info,
            shutdown_timeout=0.1,
        )

    async def _service_properties_route(
        self,
        request: aiohttp.web.Request,
    ) -> aiohttp.web.Response:
        return aiohttp.web.json_response(self.state_manager.cache_manager.cache)

    async def _web_settings_route(
        self,
        request: aiohttp.web.Request,
    ) -> aiohttp.web.Response:
        return aiohttp.web.json_response(self.web_settings)

    async def _styles_route(
        self,
        request: aiohttp.web.Request,
    ) -> aiohttp.web.FileResponse | aiohttp.web.Response:
        if self.css is not None:
            return aiohttp.web.FileResponse(self.css)

        return aiohttp.web.Response(content_type="text/css")

    def _initialise_configuration(self) -> None:
        logger.debug("Initialising web server configuration...")

        if self._generate_web_settings:
            logger.debug("Generating web settings file...")
            file_path = self._service_config_dir / "web_settings.json"

            # File does not exist, create it with default content
            file_path.parent.mkdir(
                parents=True, exist_ok=True
            )  # Ensure directory exists
            file_path.write_text(json.dumps(self.web_settings, indent=4))

    def _get_web_settings_from_file(self) -> dict[str, dict[str, Any]]:
        file_path = self._service_config_dir / "web_settings.json"
        web_settings = {}

        # File exists, read its content
        if file_path.exists():
            logger.debug(
                "Reading configuration from file '%s' ...", file_path.absolute()
            )

            web_settings = json.loads(file_path.read_text())

        return web_settings

    @property
    def web_settings(self) -> dict[str, dict[str, Any]]:
        current_web_settings = self._get_web_settings_from_file()
        for path in generate_serialized_data_paths(self.state_manager.cache_value):
            if path in current_web_settings:
                continue

            # Creating the display name by reversely looping through the path parts
            # until an item does not start with a square bracket, and putting the parts
            # back together again. This allows for display names like
            #       >>> 'dict_attr["some.dotted.key"]'
            display_name_parts: list[str] = []
            for item in parse_full_access_path(path)[::-1]:
                display_name_parts.insert(0, item)
                if not item.startswith("["):
                    break

            current_web_settings[path] = {
                "displayName": get_path_from_path_parts(display_name_parts),
                "display": True,
            }

        return current_web_settings
